
/******************************************************************************

Copyright (c) 1997, Andrew J. Weber, Jason P. Weber

Copyright information should have been included with this distribution
and may be found at ../meta/COPYRIGHT or obtained via email from weber@acm.org

******************************************************************************/

#include <bld/target.h>
#include <bld/condop.h>

/***********************************************************************
class BLD_AlternativeOperation
***********************************************************************/
BLD_Value BLD_AlternativeOperation::Perform(void) const
{
	const BLD_Value value = condition->Perform();
	if (value.IsFalse())
	{
		if (!negative.IsNull())
		{
			return negative->Perform();
		}
	}
	else
	{
		if (!affirmative.IsNull())
		{
			return affirmative->Perform();
		}
	}
	return BLD_Value::falsehood;
}

GPL_String BLD_AlternativeOperation::ToString(void) const
{
	GPL_String image;
	image.Printf("if (%s)\n{\n", condition->ToString().GetData());
	if (!affirmative.IsNull())
	{
		image.Append(affirmative->ToString());
		image.Append("\n");
	}
	image.Append("}");
	if (!negative.IsNull())
	{
		image.Append("\nelse\n{\n");
		image.Append(negative->ToString());
		image.Append("\n}");
	}
	return image;
}

/***********************************************************************
class BLD_BreakOperation
***********************************************************************/
BLD_Value BLD_BreakOperation::Perform(void) const
{
	register const unsigned long n = BLD_Target::targets.GetCount();
	if (0 < n)
	{
		if (topOnly)
		{
			BLD_Target::targets.Top()->Terminate();
		}
		else
		{
			for (register unsigned long i = n; 0 < i; --i)
			{
				BLD_Target::targets[i - 1]->Terminate();
			}
		}
	}
	return BLD_Value::truth;
}

GPL_String BLD_BreakOperation::ToString(void) const
{
	GPL_String image;
	image.Printf("%s", (topOnly ? "break" : "exit"));
	return image;
}

