/******************************************************************************

Copyright (c) 1997, Andrew J. Weber, Jason P. Weber

Copyright information should have been included with this distribution
and may be found at ../meta/COPYRIGHT or obtained via email from weber@acm.org

******************************************************************************/

#include <bld/op.h>
#include <bld/interp.h>

/***********************************************************************
class BLD_LiteralOperation
***********************************************************************/
BLD_Value BLD_LiteralOperation::Perform(void) const
{
	return string;
}

GPL_String BLD_LiteralOperation::ToString(void) const
{
	GPL_String image;
	image.Printf("'%s'", string.GetData());
	return image;
}

/***********************************************************************
class BLD_StringOperation
***********************************************************************/
BLD_Value BLD_StringOperation::Perform(void) const
{
	return BLD_CONTEXT.Evaluate(string);
}

GPL_String BLD_StringOperation::ToString(void) const
{
	GPL_String image;
	image.Printf("\"%s\"", string.GetData());
	return image;
}

/***********************************************************************
class BLD_CompositeOperation
***********************************************************************/
BLD_CompositeOperation::BLD_CompositeOperation(const GPL_Pointer<BLD_Operation> & o)
{
	ops.Resize(1);
	ops.SetElement(o, 0);
}

void BLD_CompositeOperation::Add(const GPL_Pointer<BLD_Operation> & o)
{
	ops.AppendElement(o);
}

BLD_Value BLD_CompositeOperation::Perform(void) const
{
	GPL_String string;
	BLD_Value value;
	register const unsigned long n = ops.GetCount();
	for (register unsigned long i = 0; n > i; ++i)
	{
		value = ops[i]->Perform();
		string.Append(value.GetString());
	}
//WBA_APP->Printf("BLD_CompositeOperation::Perform.string == \"%s\"", string.GetData());
	return string;
}

GPL_String BLD_CompositeOperation::ToString(void) const
{
	GPL_String image;
	register const unsigned long n = ops.GetCount();
	for (register unsigned long i = 0; n > i; ++i)
	{
		image.Append(ops[i]->ToString().GetData());
	}
	return image;
}

/***********************************************************************
class BLD_ListOperation
***********************************************************************/
BLD_Value BLD_ListOperation::Perform(void) const
{
	BLD_Value values;
	BLD_Value value;
	register const unsigned long n = ops.GetCount();
	for (register unsigned long i = 0; n > i; ++i)
	{
		value = ops[i]->Perform();
		values.AppendRange(value);
	}
//WBA_APP->Printf("BLD_ListOperation::Perform.values == \"%s\"", values.GetString().GetData());
	return values;
}

GPL_String BLD_ListOperation::ToString(void) const
{
	GPL_String image;
	register const unsigned long n = ops.GetCount();
	for (register unsigned long i = 0; n > i; ++i)
	{
		image.Join(" ", ops[i]->ToString().GetData());
	}
	return image;
}

/***********************************************************************
class BLD_BlockOperation
***********************************************************************/
BLD_Value BLD_BlockOperation::Perform(void) const
{
	BLD_Value value;
	register const unsigned long n = ops.GetCount();
	for (register unsigned long i = 0; n > i; ++i)
	{
		value = ops[i]->Perform();
	}
	return value;
}

GPL_String BLD_BlockOperation::ToString(void) const
{
	GPL_String image;
	register const unsigned long n = ops.GetCount();
	for (register unsigned long i = 0; n > i; ++i)
	{
		image.Join("\n", ops[i]->ToString().GetData());
	}
	return image;
}
