%{

/* This explicitly includes the platform-specific defines so */
/* that unistd.h can be found.                               */
#include <meta/platform.h>
#include <unistd.h>

#include <bld/op.h>
#include <bld/boolop.h>
#include <bld/relop.h>
#include <bld/fnop.h>
#include <bld/checkop.h>
#include <bld/varop.h>
#include <bld/condop.h>
#include <bld/listop.h>
#include <bld/pragmaop.h>
#include <bld/rule.h>
#include <bld/interp.h>

#define	SAFE_APPEND(x,y)	if (y) { (x)->AppendElement(y); }

extern int yyparse(void);
extern void yyerror(char *errmsg);
extern int yylex(void);

%}

%union {
	char *string;
	int integer;
	BLD_Operation *op;
	OpArray *ops;
}

%token	BLANK_SPACE NEWLINE SEMI_COLON COMMENT
%token	SINGLE_TICK DOUBLE_TICK BACK_TICK
%token	DOT COLON MINUS PLUS COMMA EQUAL PLUS_EQUAL MINUS_EQUAL
%token	LEFT_CURLY RIGHT_CURLY
%token	LEFT_PAREN RIGHT_PAREN
%token	LEFT_BRACE RIGHT_BRACE
%token	SPEC FAILSPEC PRESPEC POSTSPEC
%token	INCLUDE BUILD BUILTIN_ECHO BUILTIN_CD SYSTEM
%token	CHECK TCHECK FCHECK EXISTS DATEOF NEWEST LATEST
%token	IF ELSE ELSEIF FOREACH
%token	BEG END
%token	BREAK EXIT
%token	NAME VARIABLE LITERAL STRING COMMAND TARGET
%token	SET PRAGMA_DUMP PRAGMA_DUMPTOP PRAGMA_ENV
%token	NOT AND OR
%token	EQ NE LT LE GT GE LTLT GTGT

%left	AND OR
%left	EQ NE LT LE GT GE
%right	NOT

%type	<string>	INCLUDE BUILD BUILTIN_ECHO SYSTEM
%type	<string>	CHECK TCHECK FCHECK EXISTS DATEOF NEWEST LATEST
%type	<string>	NAME VARIABLE LITERAL TARGET
%type	<op>		STRING
%type	<integer>	assignment echo_type specifier
%type	<string>	variable
%type	<op>		value composite_list value_list command rvalue function line
%type	<op>		operand relexp boolexp if else foreach
%type	<op>		pragma set echo_dest echo cd simple complex include
%type	<op>		system build check tcheck fcheck newest latest braces
%type	<op>		parameter dateof exists target
%type	<ops>		arg_list_1 arg_list_2 arg_list param_list parameters
%type	<ops>		line_list command_list block

%start	program

%%

blank_space		:	blank_space BLANK_SPACE
				|	BLANK_SPACE
				;

value			:	LITERAL
					{
						$$ = new BLD_LiteralOperation($1);
						delete [] $1;
					}
				|	STRING
					{
						$$ = $1;
					}
				|	VARIABLE
					{
						$$ = new BLD_GetOperation($1);
						delete [] $1;
					}
				;

composite_list	:	composite_list value
					{
						((BLD_CompositeOperation *) $1)->Add($2);
						$$ = $1;
					}
				|	value value
					{
						$$ = new BLD_CompositeOperation($1);
						((BLD_CompositeOperation *) $$)->Add($2);
					}
				;

value_list		:	composite_list
					{
						$$ = $1;
					}
				|	value
					{
						$$ = $1;
					}
				;

arg_list_1		:	arg_list_1 blank_space value_list
					{
						SAFE_APPEND($1, $3);
						$$ = $1;
					}
				|	value_list
					{
						$$ = new OpArray(1);
						SAFE_APPEND($$, $1);
					}
				;

arg_list_2		:	arg_list_2 value_list blank_space
					{
						SAFE_APPEND($1, $2);
						$$ = $1;
					}
				|	blank_space
					{
						$$ = new OpArray();
					}
				;

arg_list		:	arg_list_1
					{
						$$ = $1;
					}
				|	arg_list_1 blank_space
					{
						$$ = $1;
					}
				|	arg_list_2 value_list
					{
						SAFE_APPEND($1, $2);
						$$ = $1;
					}
				|	arg_list_2
					{
						$$ = $1;
					}
				;

param_separator	:	BLANK_SPACE COMMA BLANK_SPACE
				|	BLANK_SPACE COMMA
				|	COMMA BLANK_SPACE 
				|	COMMA
				;

param_list		:	param_list param_separator value_list
					{
						SAFE_APPEND($1, $3);
						$$ = $1;
					}
				|	value_list
					{
						$$ = new OpArray(1);
						SAFE_APPEND($$, $1);
					}
				;

left_paren		:	BLANK_SPACE LEFT_PAREN
				|	LEFT_PAREN
				;

right_paren		:	BLANK_SPACE RIGHT_PAREN
				|	RIGHT_PAREN
				;

parameter		:	left_paren BLANK_SPACE value_list right_paren
					{
						$$ = $3;
					}
				|	left_paren value_list right_paren
					{
						$$ = $2;
					}
				;

parameters		:	left_paren BLANK_SPACE param_list right_paren
					{
						$$ = $3;
					}
				|	left_paren param_list right_paren
					{
						$$ = $2;
					}
				;

system			:	SYSTEM parameters 
					{
						if ((OpArray *) 0 == $2)
						{
							$$ = new BLD_SystemOperation();
						}
						else
						{
							$$ = new BLD_SystemOperation(*($2));
							delete $2;
						}
					}
				;

build			:	BUILD parameters 
					{
						if ((OpArray *) 0 == $2)
						{
							$$ = new BLD_BuildOperation();
						}
						else
						{
							$$ = new BLD_BuildOperation(*($2));
							delete $2;
						}
					}
				;

check			:	CHECK parameters 
					{
						if ((OpArray *) 0 == $2)
						{
							$$ = new BLD_CheckOperation();
						}
						else
						{
							$$ = new BLD_CheckOperation(*($2));
							delete $2;
						}
					}
				;

tcheck			:	TCHECK parameters 
					{
						if ((OpArray *) 0 == $2)
						{
							$$ = new BLD_TargetCheckOperation();
						}
						else
						{
							$$ = new BLD_TargetCheckOperation(*($2));
							delete $2;
						}
					}
				;

fcheck			:	FCHECK parameters 
					{
						if ((OpArray *) 0 == $2)
						{
							$$ = new BLD_FileCheckOperation();
						}
						else
						{
							$$ = new BLD_FileCheckOperation(*($2));
							delete $2;
						}
					}
				;

exists			:	EXISTS parameter
					{
						$$ = new BLD_ExistOperation($2);
					}
				;

dateof			:	DATEOF parameter
					{
						$$ = new BLD_DateOperation($2);
					}
				;

newest			:	NEWEST parameters 
					{
						if ((OpArray *) 0 == $2)
						{
							$$ = new BLD_DateSortOperation();
						}
						else
						{
							$$ = new BLD_DateSortOperation(*($2), TRUE);
							delete $2;
						}
					}
				;

latest			:	LATEST parameters 
					{
						if ((OpArray *) 0 == $2)
						{
							$$ = new BLD_DateSortOperation();
						}
						else
						{
							$$ = new BLD_DateSortOperation(*($2), FALSE);
							delete $2;
						}
					}
				;

braces			:	LEFT_BRACE arg_list RIGHT_BRACE
					{
						if ((OpArray *) 0 == $2)
						{
							$$ = new BLD_LiteralOperation
										(BLD_Value::falsehood);
						}
						else
						{
							$$ = new BLD_EqualOperation
								(new BLD_LiteralOperation
									(BLD_BuildOperation::built.GetData()),
								new BLD_CheckOperation(*($2)));
							delete $2;
						}
					}
				;

function		:	system
					{
						$$ = $1;
					}
				|	build
					{
						$$ = $1;
					}
				|	check
					{
						$$ = $1;
					}
				|	tcheck
					{
						$$ = $1;
					}
				|	fcheck
					{
						$$ = $1;
					}
				|	exists
					{
						$$ = $1;
					}
				|	dateof
					{
						$$ = $1;
					}
				|	newest
					{
						$$ = $1;
					}
				|	latest
					{
						$$ = $1;
					}
				|	braces
					{
						$$ = $1;
					}
				;

operand			:	BLANK_SPACE value_list
					{
						$$ = $2;
					}
				|	value_list
					{
						$$ = $1;
					}
				|	BLANK_SPACE function
					{
						$$ = $2;
					}
				|	function
					{
						$$ = $1;
					}
				;

relexp			:	operand EQ operand
					{
						$$ = new BLD_EqualOperation($1, $3);
					}
				|	operand NE operand
					{
						$$ = new BLD_NotEqualOperation($1, $3);
					}
				|	operand LT operand
					{
						$$ = new BLD_LessOperation($1, $3);
					}
				|	operand LE operand
					{
						$$ = new BLD_LessEqualOperation($1, $3);
					}
				|	operand GT operand
					{
						$$ = new BLD_GreaterOperation($1, $3);
					}
				|	operand GE operand
					{
						$$ = new BLD_GreaterEqualOperation($1, $3);
					}
				;

boolexp			:	boolexp AND boolexp
					{
						$$ = new BLD_AndOperation($1, $3);
					}
				|	boolexp OR boolexp
					{
						$$ = new BLD_OrOperation($1, $3);
					}
				|	NOT boolexp
					{
						$$ = new BLD_NotOperation($2);
					}
				|	relexp
					{
						$$ = $1;
					}
				|	operand
					{
						$$ = $1;
					}
				|	left_paren boolexp right_paren
					{
						$$ = $2;
					}
				;

variable		:	BLANK_SPACE VARIABLE
					{
						$$ = $2;
					}
				|	VARIABLE
					{
						$$ = $1;
					}
				;

rvalue			:	function
					{
						$$ = $1;
					}
				|	left_paren boolexp right_paren
					{
						$$ = $2;
					}
				|	arg_list
					{
						if ((OpArray *) 0 == $1)
						{
							$$ = (BLD_Operation *) GPL_BAD_ADDRESS;
						}
						else
						{
							$$ = new BLD_ListOperation(*($1));
							delete $1;
						}
					}
				;

assignment		:	EQUAL
					{
						$$ = BLD_SET_OVERWRITE;
					}
				|	PLUS_EQUAL
					{
						$$ = BLD_SET_APPEND;
					}
				|	MINUS_EQUAL
					{
						$$ = BLD_SET_PREPEND;
					}
				;

echo_dest		:	BLANK_SPACE value_list
					{
						$$ = $2;
					}
				|	value_list
					{
						$$ = $1;
					}
				;

echo_type		:	GTGT
					{
						$$ = TRUE;
					}
				|	GT
					{
						$$ = FALSE;
					}
				;

echo			:	BUILTIN_ECHO BLANK_SPACE arg_list
					{
						if ((OpArray *) 0 == $3)
						{
							$$ = new BLD_EchoOperation();
						}
						else
						{
							$$ = new BLD_EchoOperation(*($3));
							delete $3;
						}
					}
				|	BUILTIN_ECHO BLANK_SPACE arg_list echo_type echo_dest
					{
						if ((OpArray *) 0 == $3)
						{
							$$ = new BLD_FileEchoOperation($5, $4);
						}
						else
						{
							$$ = new BLD_FileEchoOperation(*($3), $5, $4);
							delete $3;
						}
					}
				;

cd				:	BUILTIN_CD BLANK_SPACE value_list
					{
						$$ = new BLD_ChangeDirectoryOperation($3);
					}
				|	BUILTIN_CD value_list
					{
						$$ = new BLD_ChangeDirectoryOperation($2);
					}
				;

pragma			:	PRAGMA_DUMP
					{
						$$ = new BLD_PragmaDumpOperation(FALSE);
					}
				|	PRAGMA_DUMPTOP
					{
						$$ = new BLD_PragmaDumpOperation(TRUE);
					}
				|	PRAGMA_ENV
					{
						$$ = new BLD_PragmaEnvOperation();
					}
				;

set				:	SET variable assignment rvalue
					{
						$$ = new BLD_SetOperation((BLD_SetType) $3, $2, $4);
						delete [] $2;
					}
				|	SET variable assignment
					{
						$$ = new BLD_SetOperation((BLD_SetType) $3, $2,
													new BLD_Operation());
						delete [] $2;
					}
				;

left_curly		:	BLANK_SPACE LEFT_CURLY
				|	LEFT_CURLY
				;

right_curly		:	BLANK_SPACE RIGHT_CURLY
				|	RIGHT_CURLY
				;

block			:	left_curly command_list right_curly
					{
						$$ = $2;
					}
				;

if				:	IF left_paren boolexp right_paren block else
					{
						$$ = new BLD_AlternativeOperation($3,
									new BLD_BlockOperation(*($5)), $6);
						delete $5;
					}
				|	IF left_paren boolexp right_paren block
					{
						$$ = new BLD_AlternativeOperation($3,
									new BLD_BlockOperation(*($5)));
						delete $5;
					}
				;

else			:	ELSEIF left_paren boolexp right_paren block else
					{
						$$ = new BLD_AlternativeOperation($3,
									new BLD_BlockOperation(*($5)), $6);
						delete $5;
					}
				|	ELSEIF left_paren boolexp right_paren block
					{
						$$ = new BLD_AlternativeOperation($3,
									new BLD_BlockOperation(*($5)));
						delete $5;
					}
				|	ELSE block
					{
						$$ = new BLD_BlockOperation(*($2));
						delete $2;
					}
				;

foreach			:	FOREACH blank_space LITERAL left_paren operand right_paren block
					{
						$$ = new BLD_ForEachOperation($3, $5, *($7));
						delete [] $3;
						delete $7;
					}
				;

include			:	INCLUDE BLANK_SPACE value_list
					{
						$$ = new BLD_IncludeOperation($3);
					}
				|	INCLUDE value_list
					{
						$$ = new BLD_IncludeOperation($2);
					}
				;

specifier		:	PRESPEC
					{
						$$ = BLD_RULE_PREPEND_OPERATIONS;
					}
				|	FAILSPEC
					{
						$$ = BLD_RULE_FAIL_OPERATIONS;
					}
				|	POSTSPEC
					{
						$$ = BLD_RULE_APPEND_OPERATIONS;
					}
				|	SPEC
					{
						$$ = BLD_RULE_OVERWRITE_OPERATIONS;
					}
				;

target			:	TARGET specifier block
					{
						$$ = new BLD_SpecificationOperation($1, (const BLD_RuleType) $2, *($3));
						delete [] $1;
						delete $3;
					}
				;

simple			:	COMMENT
					{
						$$ = (BLD_Operation *) GPL_BAD_ADDRESS;
					}
				|	pragma
					{
						$$ = $1;
					}
				|	echo
					{
						$$ = $1;
					}
				|	cd
					{
						$$ = $1;
					}
				|	set
					{
						$$ = $1;
					}
				|	function
					{
						$$ = $1;
					}
				|	value
					{
						OpArray ops;
						ops.Resize(1);
						ops.SetElement($1, 0);
						$$ = new BLD_SystemOperation(ops);
					}
				|	value arg_list
					{
						OpArray ops;
						ops.Resize(($2)->GetCount() + 1);
						ops.SetElement($1, 0);
						ops.AppendRange(*($2));
						$$ = new BLD_SystemOperation(ops);
						delete $2;
					}
				|	if
					{
						$$ = $1;
					}
				|	foreach
					{
						$$ = $1;
					}
				|	BREAK
					{
						$$ = new BLD_BreakOperation(TRUE);
					}
				|	EXIT
					{
						$$ = new BLD_BreakOperation(FALSE);
					}
				|	include
					{
						$$ = $1;
					}
				|	target
					{
						$$ = $1;
					}
				;

command_separator	:	SEMI_COLON
					|	NEWLINE
					;

command			:	BLANK_SPACE simple command_separator
					{
						$$ = $2;
					}
				|	simple command_separator
					{
						$$ = $1;
					}
				|	BLANK_SPACE command_separator
					{
						$$ = (BLD_Operation *) GPL_BAD_ADDRESS;
					}
				|	command_separator
					{
						$$ = (BLD_Operation *) GPL_BAD_ADDRESS;
					}
				;

command_list	:	command_list command
					{
						SAFE_APPEND($1, $2);
						$$ = $1;
					}
				|	command
					{
						$$ = new OpArray(1);
						SAFE_APPEND($$, $1);
					}
				;

complex			:	simple
					{
						$$ = $1;
					}
				|	BEG NAME
					{
						$$ = new BLD_BeginScopeOperation($2);
						delete [] $2;
					}
				|	END
					{
						$$ = new BLD_EndScopeOperation();
					}
				;

line			:	BLANK_SPACE complex command_separator
					{
						$$ = $2;
					}
				|	complex command_separator
					{
						$$ = $1;
					}
				|	BLANK_SPACE command_separator
					{
						$$ = (BLD_Operation *) GPL_BAD_ADDRESS;
					}
				|	command_separator
					{
						$$ = (BLD_Operation *) GPL_BAD_ADDRESS;
					}
				;

line_list		:	line_list line
					{
						SAFE_APPEND($1, $2);
						$$ = $1;
					}
				|	line
					{
						$$ = new OpArray(1);
						SAFE_APPEND($$, $1);
					}
				;

program			:	line_list
					{
						*BLD_Interpreter::operations = *($1);
						delete $1;
					}
				;

%%

#if OSD_OS==OSD_HPUX
void *alloca(size_t size)
{
	return malloc(size);
}
#endif /* OSD_OS==OSD_HPUX */
