#!/bin/sh

append()
{
	if [ -n "$NOTE" ]
	then
		NOTE="${NOTE}\n"
	fi
	NOTE="${NOTE}$1"
}

# Set up the defaults.
USAGE="Usage: $0 [-f|-i|-x] -m module [notes...]"
TYPE=FEATURE
DATE=0

# Parse the command-line arguments.
while getopts \?dfim:x OPT
do
	case $OPT in
	d)	DATE=1 ;;
	f)	TYPE=FEATURE ;;
	i)	TYPE=INTERFACE ;;
	m)	MODULE=$OPTARG ;;
	x)	TYPE=FIX ;;
	\?)	echo $USAGE
		exit 1 ;;
	esac
done

# Module name must be provided.
if [ -z "$MODULE" ]
then
	echo $USAGE
	exit 1
fi

# Convert the module name to upper-case.
MODULE=`echo $MODULE | tr "[a-z]" "[A-Z]"`

# Add any notes from the command-line.
shift `expr $OPTIND - 1`
while [ 0 -lt $# ]
do
	append "$1"
	shift
done

# If there are still no notes...
if [ -z "$NOTE" ]
then
	# If standard input comes from a terminal...
	if [ -t 0 ]
	then
		echo "Enter note:"
	fi

	# Read the note.
	read LINE
	while [ 0 -eq $? ]
	do
		append "$LINE"
		read LINE
	done
fi

# Go to the meta directory.
META_DIRECTORY=`dirname $0`
cd $META_DIRECTORY

# Update RELEASE_NOTES to minimize conflicts.
cvs update RELEASE_NOTES

# Add the release note.
if [ 1 -eq $DATE ]
then
	echo "\n$MODULE $TYPE `date '+%Y.%m.%d %H:%M:%S'`" >> RELEASE_NOTES
else
	echo "\n$MODULE $TYPE" >> RELEASE_NOTES
fi
echo "$NOTE" >> RELEASE_NOTES

# Propagate changes.
cvs commit -m "$MODULE $TYPE" RELEASE_NOTES

# Success!
exit 0
