
/******************************************************************************

COPYRIGHT_HEADER

******************************************************************************/

#include <mte/mte.h>

long MTE_Task::total_starts = 0;
long MTE_Task::active_threads = 0;
MTE_Mutex track_lock;

MTE_Task::MTE_Task()
{
	state = MTE_TASK_READY;
#if MTE_IMPL==MTE_SPROC
#elif MTE_IMPL==MTE_PTHREAD
	pthread_attr_init(&attr);
	pthread_attr_setscope(&attr, PTHREAD_SCOPE_PROCESS);
#elif MTE_IMPL==MTE_WIN32
#else
#endif
	SetDetach(FALSE);
}

MTE_Task::~MTE_Task()
{
#if MTE_IMPL==MTE_SPROC
#elif MTE_IMPL==MTE_PTHREAD
	pthread_attr_destroy(&attr);
#elif MTE_IMPL==MTE_WIN32
#else
#endif
}

void MTE_Task::SetDetach(GPL_Boolean setting)
{
#if MTE_IMPL==MTE_SPROC
#elif MTE_IMPL==MTE_PTHREAD
	if(setting)
	{
		pthread_attr_setdetachstate( &attr, PTHREAD_CREATE_DETACHED);
	}
	else
	{
		pthread_attr_setdetachstate( &attr, PTHREAD_CREATE_JOINABLE);
	}
#elif MTE_IMPL==MTE_WIN32
#else
#endif
}

void MTE_Task::SetKillable(GPL_Boolean setting)
{
#if MTE_IMPL==MTE_SPROC
#elif MTE_IMPL==MTE_PTHREAD
	int dummy;
	if(setting)
	{
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, &dummy);
		pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, &dummy);
	}
	else
	{
		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &dummy);
		pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, &dummy);
	}
#elif MTE_IMPL==MTE_WIN32
#else
#endif
}

#if MTE_IMPL==MTE_SPROC
void mteRelayCallbackSproc(void *argument)
{
	MTE_Task *task;
	task = (MTE_Task *)argument;
	task->Work();
	return;
}
#elif MTE_IMPL==MTE_PTHREAD
void *mteRelayCallbackPthread(void *argument)
{
	MTE_Task *task;
	task = (MTE_Task *)argument;
	task->Work();
	return NULL;
}
#elif MTE_IMPL==MTE_WIN32
void mteRelayCallbackWin32(void *argument)
{
	MTE_Task *task;
	task = (MTE_Task *)argument;
	task->Work();
	return;
}
#else
#endif

GPL_Boolean MTE_Task::Start()
{
	track_lock.Lock();
	total_starts++;
	active_threads++;
	track_lock.Unlock();
	state = MTE_TASK_READY;
#if MTE_IMPL==MTE_SPROC
	thread = sproc(&mteRelayCallbackSproc,PR_SADDR,(void *)this);
	if(thread == -1)
	{
		return FALSE;
	}
	else
	{
		return TRUE;
	}
#elif MTE_IMPL==MTE_PTHREAD
	int n;
	n = pthread_create( &thread, NULL,&mteRelayCallbackPthread, (void *)this);
	if(n)
	{
		wbaLog("MTE ERROR", "could not create pthread");
		return FALSE;
	}
	return TRUE;
#elif MTE_IMPL==MTE_WIN32
	thread = CreateThread(
		NULL,
		0,
		(LPTHREAD_START_ROUTINE) &mteRelayCallbackWin32,
		(LPVOID) this,
		0,
		&dwThreadId);
	if(thread)
	{
		return TRUE;
	}
	return FALSE;
#else
	Work();
	return TRUE;
#endif
}

GPL_Boolean MTE_Task::Wait()
{
#if MTE_IMPL==MTE_SPROC
	pid_t pid;
	int status;

	long wid=(long)thread;
	if(wid<1) wid= -1;

	/* POSIX child wait */
	pid=waitpid(wid, &status, NULL);

	if(pid>0)
	{
		return TRUE;
	}
	return FALSE;
#elif MTE_IMPL==MTE_PTHREAD
	int n;
	n = pthread_join( thread, NULL );
	if(n)
	{
		return FALSE;
	}
	return TRUE;
#elif MTE_IMPL==MTE_WIN32
	WaitForSingleObject(thread, INFINITE);
	return TRUE;
#else
	return TRUE;
#endif
}

GPL_Boolean MTE_Task::Kill()
{
#if MTE_IMPL==MTE_SPROC
	int status;

	status = sigsend(P_PID, (pid_t)thread, SIGTERM);
	if(!status) { return TRUE; }
	return FALSE;
#elif MTE_IMPL==MTE_PTHREAD
	int n;
	n = pthread_cancel( thread );
	if(n) { return FALSE; }
	return TRUE;
#elif MTE_IMPL==MTE_WIN32
	BOOL status;
	status = TerminateThread(thread, 0);
	if(status){ return TRUE; }
	return FALSE;
#else
	return TRUE;
#endif
}

void MTE_Task::Work()
{
	wbaLog("MTE LOG", "default MTE_Task::Work()");
}

void MTE_Task::Stop()
{
	track_lock.Lock();
	active_threads--;
	track_lock.Unlock();
#if MTE_IMPL==MTE_SPROC
	exit(0);
#elif MTE_IMPL==MTE_PTHREAD
	pthread_exit(NULL);
#elif MTE_IMPL==MTE_WIN32
	ExitThread(0);
#else
#endif
}

void MTE_Task::Done()
{
	state = MTE_TASK_DONE;
}


