/******************************************************************************

Coral Tree spa/integrate.cc
Copyright (C) 1998 Infinite Monkeys Incorporated

This program is free software; you can redistribute it and/or modify
it under the terms of the Artistic License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Artistic License for more details.

You should have received a copy of the Artistic License along with
this program (see meta/ART_LICE); if not, write to coral@imonk.com.

******************************************************************************/

#include <spa/spa.h>

SPA_Integrate::SPA_Integrate()
{
	divisions = 2;
}

SPA_Integrate::~SPA_Integrate()
{
}

double SPA_Integrate::Trapezoid(double start, double end)
{
	double y,t;
	double gap;
	gap = fabs(end - start);
	y = Function(start);
	t = Function(end);
	return (y + t)*gap/2.0;
}

double SPA_Integrate::Rectangle(double start, double end)
{
	double mid, gap;
	gap = fabs(end - start);
	mid = start+mid;
	return gap*Function(mid);

}

double SPA_Integrate::Quadrature(double start,double end,double tol,long depth,double a0)
{
//fprintf(stderr,"-- %g %g\n", start, end);
	double a1, a2, a3, midpoint, err, y2, y3, x2, x3;
	if(depth)
	{
		a1 = a0;
	}
	else
	{
		a1 = Trapezoid(start, end);
	}
	midpoint = start + (end-start)/2.0;
	double gap;
	gap = midpoint - start;
	x2 = start+gap/2.0;
	x3 = midpoint+gap/2.0;
	a2 = gap*Function(x2);
	a3 = gap*Function(x3);
	//a2 = Trapezoid(start,midpoint);
	//a3 = Trapezoid(midpoint,end);
	//double b1, b2;
	//b1 = Function(start);
	//b2 = Function(end);
	err = tol;
	if(fabs((x3-x2)/x2) > tol && !(fabs(x2)<1e-100 && fabs(x3)<1e-100))
	//if( (fabs(a2+a3) > tol) ||
	//	(fabs((b2-b1)/b1) > tol && !(fabs(b1)<1e-100 && fabs(b2)<1e-100)) ||
	//	(fabs(a2+a3-a1) > tol) ||
	//	(fabs(end-start) > 1e100*tol)
	//)
	//if(fabs(a2+a3) > tol)
	//if(fabs((b2-b1)/b1) > tol && !(fabs(b1)<1e-100 && fabs(b2)<1e-100))
	//if(fabs(a2+a3-a1) > tol)
	{
//fprintf(stderr,"%g %g %d %g %g %g %g\n", start, midpoint, depth, a1, a2, fabs(a2+a3-a1), err);
		a1 = a2+a3;
		//err = err/2.0;
		a2 = Quadrature(start, midpoint, err, depth+1, a2);
		a3 = Quadrature(midpoint, end, err, depth+1, a3);
		err = err/2.0;
	}
	return a2+a3;
}

double SPA_Integrate::Trapezoidx(double start, double end, double maxy)
{
//fprintf(stderr,"start %.10f end %.10f\n", start, end);
	double area;
	if( 1 == divisions )
	{
		area = (Function(start) + Function(end))*fabs(end - start)/2.0;
	}
	else
	{
		double gap = fabs(end - start)/(double)divisions;
		double x = (start<end)?start:end;
		double y,t,l;
		double mid;
		area = 0.0;
		y = Function(x);
//fprintf(stderr, "y %f x %f\n", y, x);
		for(long i=0;i < divisions;i++)
		{
			t = Function(x+gap);
			mid = Function(x+gap/2.0);
//fprintf(stderr,"fabs(t-y) %f(%f - %f)\n", fabs(t-y),t,y, x, gap);
#if 1
			l = fabs(t - y);
//fprintf(stderr,"fabs(l/y) %f\n", fabs(l/y));
if(fabs(l/y) == 1.0)
{
	//fprintf(stderr,"fabs(l/y) %f t %.12f y %.12f\n", fabs(l/y), t, y);
}
			if(fabs(l/y) > maxy && !(fabs(y)<1e-100 && fabs(t)<1e-100))
			//if(fabs(l/y) > maxy)
			{
				area += Trapezoidx(x, x+gap, maxy);
			}
			else
			{
				area += (y + t)*gap/2.0;
				//area += mid*gap;
			}
//fprintf(stderr,"area %.10f %.10f to %.10f %f %f\n", area, x, x+gap, y, t);
#else
			area += (y + t)*gap/2.0;
#endif
			x += gap;
			y = t;
		}
	}
//fprintf(stderr,"area %.16f\n", area);
	return area;
}

double SPA_Integrate::Trapezoid2(double start, double end, double maxy)
{
//fprintf(stderr,"start %.10f end %.10f\n", start, end);
	double area;
	if( 1 == divisions )
	{
		area = (Function(start) + Function(end))*fabs(end - start)/2.0;
	}
	else
	{
		double gap = fabs(end - start)/(double)divisions;
		double x = (start<end)?start:end;
		double y,t;
		area = 0.0;
		y = Function(x);
//fprintf(stderr, "y %f x %f\n", y, x);
		for(long i=0;i < divisions;i++)
		{
			t = Function(x+gap);
//fprintf(stderr,"fabs(t-y) %f(%f - %f)\n", fabs(t-y),t,y, x, gap);
#if 1
			if(fabs(t-y) > maxy)
			{
				area += Trapezoidx(x, x+gap, maxy);
			}
			else
			{
				area += (y + t)*gap/2.0;
			}
//fprintf(stderr,"area %.10f %.10f to %.10f\n", area, x, x+gap);
#else
			area += (y + t)*gap/2.0;
#endif
			x += gap;
			y = t;
		}
	}
//fprintf(stderr,"area %.16f\n", area);
	return area;
}



