/******************************************************************************

Coral Tree spa/spline.cc
Copyright (C) 1998 Infinite Monkeys Incorporated

This program is free software; you can redistribute it and/or modify
it under the terms of the Artistic License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Artistic License for more details.

You should have received a copy of the Artistic License along with
this program (see meta/ART_LICE); if not, write to coral@imonk.com.

******************************************************************************/



#include <spa/spa.h>



#define	SPA_CARDINALSCALE 0.5	// (0.5) tangent length in cardinal spline



/******************************************************************************
float   SPA_Spline::Cardinal2(float t,float p0,float p1,float p2,float p3)

	static

	letting f(-1) = p0
			f( 0) = p1
			f( 1) = p2
			f( 2) = p3
	
	returns f(t)  for  0.0 < t < 1.0
		
	uses Cardinal spline interpolation of a curve (2D)

	value   = [t^3 t^2 t 1] / -a 2-a a-2   a \  / p0 \
							| 2a a-3 3-2a -a |  | p1 |
							| -a  0   a    0 |  | p2 |
							\  0  1   0    0 /  \ p3 /

			= p0 (-at^3 + 2at^2 - at)
			+ p1 (2t^3 - at^3 - 3t^2 + at^2 + 1)
			+ p2 (-2t^3 + at^3 + 3t^2 - 2at^2 + at)
			+ p3 (at^3 - at^2)

	The Cardinal spline guarantees control-point intersection.

******************************************************************************/
float SPA_Spline::Cardinal2D(float t,float p0,float p1,float p2,float p3)
	{
	float t2,t3,at,at2,at3,dt3,dat2,tt2;	/* intermediate values */
	float value;							/* resultant splined value */

	t2=t*t;
	t3=t2*t;
	at=SPA_CARDINALSCALE*t;
	at2=at*t;
	at3=at2*t;
	dt3=2.0*t3;
	dat2=2.0*at2;
	tt2=3.0*t2;

	value=p0*(-at3+dat2-at)+p1*(dt3-at3+at2-tt2+1)+p2*(at3-dt3+tt2-dat2+at)+p3*(at3-at2);

	return value;
	}



/******************************************************************************
float   SPA_Spline::Cardinal3b(float x,float y,float p[4][4])

	static

	XY-separated version

	simple implementation of surface spline using multiple calls
		to Cardinal2()

	apparently equivalent to true 3D Cardinal, but twice as fast

	Note: It would appear that the Cardinal spline is x-y separable.  I am
		not sure why, but the measured results are the same.
******************************************************************************/
float SPA_Spline::Cardinal3D(float x,float y,float p[4][4])
	{
	float p1,p2,p3,p4;	// 4 values of y along the given value of x
	float val;			// resulting value to be returned

	p1=Cardinal2D(x,p[0][0],p[0][1],p[0][2],p[0][3]);
	p2=Cardinal2D(x,p[1][0],p[1][1],p[1][2],p[1][3]);
	p3=Cardinal2D(x,p[2][0],p[2][1],p[2][2],p[2][3]);
	p4=Cardinal2D(x,p[3][0],p[3][1],p[3][2],p[3][3]);

	val=Cardinal2D(y,p1,p2,p3,p4);

	return val;
	}



/******************************************************************************
float   SPA_Spline::FlatAverage(float x,float y,float p0,float p1,float p2,float p3)

	computes value at a point within values of 4 surrounding points

	simple x-y linear interpolation
******************************************************************************/
float SPA_Spline::FlatAverage(float x,float y,float p0,float p1,float p2,float p3)
	{
	float y1,y2;	/* y values at x==0 and x==1 */
	float result;	/* resulting value to be returned */

	y1=p0+(p3-p0)*y;
	y2=p1+(p2-p1)*y;

	result=y1+(y2-y1)*x;

	return result;
	}

